obj = ScanLMM(
    phenoFile = 'examples/toydata.phe',
    
    phenoTestFile = 'examples/toydata.phe',
    
    bedFile = 'examples/toydata',
    outFileLMM = 'examples/LmmGWAS.out.txt',

    #fastlmmPath = 'C:/Users/t-baraki/Users/t-baraki/adaptivelmm/software/FaSTLMM.205.Win/Cpp_MKL',
    fastlmmPath = r"C:\Users\lippert\Projects\FaST-LMM_old\FastLmm\CPP\x64\MKL-Release",
    excludeByPosition = 10,
    logdelta = 2,
    #topKbyLinReg = 10,

    outFileLM = 'examples/LmGWAS.out.txt',
    #outFileLinReg = 'examples/LmmGWAS1E1.snps.txt',
    #extractSim = 'examples/LmmGWAS1E3.snps.txt'

)
